from dockwrkr import (Command)


class Recreate(Command):

    def getShellOptions(self, optparser):
        optparser.add_option("-a", "--all", dest="allc",
                             help="Apply to defined containers", default=False, action="store_true")
        optparser.add_option("-t", "--time", dest="time",
                             help="Seconds to wait before sending SIGKILL", default=10)
        return optparser

    def getUsage(self):
        return "dockwrkr recreate [options] CONTAINER..."

    def getHelpTitle(self):
        return "Stop, Remove and Start the specified container(s)"

    def main(self):
        containers = self.args
        if not len(self.args) > 0 and not self.options.allc:
            return self.exitWithHelp("Please provide a container or use -a for all containers.")
        return self.core.recreate(self.args, all=self.getOption('allc'), time=self.getOption('time')) \
            .catch(self.exitError)



from dockwrkr import (Command)


class Help(Command):

    def __init__(self):
        super(Help, self).__init__()
        self.autoInitCore = False

    def getUsage(self):
        return "dockwrkr help [command?]"

    def getHelpTitle(self):
        return "Print help for a specific command"

    def getShellOptions(self, optparser):
        return optparser

    def main(self):
        if len(self.args) > 0:
            parent = self.parent
            for commandName in self.args:
                command = parent.getCommand(commandName)
                command.initialize()
                parent = command
        else:
            command = self.parent
        return command.exitHelp()

    def getInputCommand(self):
        name = self.args[0]
        return name



import unittest
import tempfile
import sys
import pprint
from dockwrkr.shell import Shell


class TestBase(unittest.TestCase):

    tempFiles = []
    tempDirs = []

    @classmethod
    def addTemporaryFile(cls):
        tempFile = tempfile.NamedTemporaryFile()
        cls.tempFiles.append(tempFile)
        return tempFile

    @classmethod
    def addTemporaryDir(cls):
        tempDir = tempfile.mkdtemp()
        cls.tempDirs.append(tempDir)
        return tempDir

    @classmethod
    def clearTemporaryDirs(cls):
        for x in cls.tempDirs:
            Shell.nukeDirectory(x)

    @classmethod
    def clearTemporaryFiles(cls):
        for x in cls.tempFiles:
            x.close()

    @staticmethod
    def raiser(e):
        raise e



from dockwrkr import (Command)


class Reset(Command):

    def __init__(self):
        super(Reset, self).__init__()
        self.autoInitCore = False

    def getShellOptions(self, optparser):
        optparser.add_option("-t", "--time", dest="time",
                             help="Seconds to wait before sending SIGKILL", default=10)
        return optparser

    def getUsage(self):
        return "dockwrkr reset [options]"

    def getHelpTitle(self):
        return "Reset container managed by dockwrkr (stop/remove)"

    def main(self):
        return self.core.reset(time=self.getOption('time')).catch(self.exitError)



import os
import sys
import logging
import re
from dockwrkr.monads import *
from dockwrkr.logs import *
from dockwrkr.exceptions import *
from dockwrkr.shell import Shell
from dockwrkr.utils import (readYAML, mergeDict, ensureList,
                            dateToAgo, walkUpForFile, writeToFile, expandLocalPath)
import dockwrkr.docker as docker

logger = logging.getLogger(__name__)


class Core(object):

    def __init__(self):
        self.options = {}
        self.configFile = None
        self.initialized = False
        self.config = {}
        return

    def initialize(self):
        if self.initialized:
            return OK(None)
        return self.loadConfig().then(defer(self.setInitialized, b=True))

    def setInitialized(self, b):
        self.initialized = b

    def loadConfig(self):
        return self.readConfigFile() >> self.setConfig

    def findConfigFile(self):
        if self.configFile:
            return OK(self.configFile)

        configFile = walkUpForFile(os.getcwd(), "dockwrkr.yml")
        if not configFile:
            return Fail(ConfigFileNotFound("Could not locate config file: dockwrkr.yml"))
        self.configFile = configFile
        return OK(configFile)

    def readConfigFile(self):
        return self.findConfigFile().bind(lambda f: Try.attempt(readYAML, f))

    def setConfig(self, config):
        mergeDict(self.config, config)
        return OK(self)

    def getRegistries(self):
        regs = self.config.get('registries', {})
        return regs

    def getNetworks(self):
        networks = self.config.get('networks', {})
        return networks

    def getDefinedServices(self, func=None, all=False):
        config_name = 'services'
        if not self.config.get(config_name):
            config_name = 'containers'
        return self.getDefinedContainers(config_name, func, all)

    def getDefinedJobs(self):
        return self.getDefinedContainers('jobs')

    def getDefinedContainers(self, configName='containers', func=None, all=False):
        graph = []
        containers = self.config.get(configName, {})

        for (container, container_config) in containers.items():
            if func == self.__start and all and not container_config.get('autostart', True):
                continue

            node = self.getContainerDependencies(container)
            graph.append(node)

        def resolveDependencies(node, resolved):
            for dep in node['deps']:
                depnode = self.getContainerDependencies(dep)
                if depnode['name'] not in resolved:
                    resolveDependencies(depnode, resolved)
            if node['name'] not in resolved:
                resolved.append(node['name'])
        resolved = []
        for node in graph:
            resolveDependencies(node, resolved)

        return resolved

    def getContainerDependencies(self, container):
        node = {}
        node['name'] = container
        deps = []
        config = self.getContainerConfig(container)
        if not config:
            raise InvalidContainerError("Container named '%s' is listed as a dependency but it is never defined." % container)
        if "link" in config:
            for link in ensureList(config['link']):
                deps.append(link.partition(':')[0])
        node['deps'] = deps
        return node

    def getBasePath(self):
        return os.path.dirname(self.configFile)

    def getContainerConfig(self, container):
        config = self.getServiceConfig(container)
        if config:
            return config
        return self.getJobConfig(container)

    def getServiceConfig(self, container):
        return self.config.get('services', self.config.get('containers', {})).get(container)

    def getJobConfig(self, container):
        return self.config.get('jobs', {}).get(container)

    def getContainerImage(self, container):
        conf = self.getContainerConfig(container)
        return conf.get('image', None)

    # Commands

    def readOrderedContainers(self, containers=[]):
        try:
            defined = self.getDefinedServices()
        except Exception as e:
            return Fail(e)
        missing = [x for x in containers if x not in defined]
        ordered = [x for x in defined if x in containers]
        if missing:
            return Fail(InvalidContainerError("Container '%s' not defined." % ' '.join(missing)))
        return OK(ordered)

    def create(self, containers=[], all=False):
        return self.__command(self.__create, containers=containers, all=all)

    def start(self, containers=[], all=False):
        return self.__command(self.__start, containers=containers, all=all)

    def stop(self, containers=[], all=False, time=docker.DOCKER_STOP_TIME):
        return self.__command(self.__stop, containers=containers, all=all, time=time)

    def remove(self, containers=[], all=False, time=docker.DOCKER_STOP_TIME, force=False):
        return self.__command(self.__remove, containers=containers, all=all, time=time, force=force)

    def restart(self, containers=[], all=False, time=docker.DOCKER_STOP_TIME):
        return self.__command(self.__restart, containers=containers, all=all, time=time)

    def stats(self, containers=[]):
        if not containers:
            try:
                containers = self.getDefinedServices()
            except Exception as e:
                return Fail(e)
        return self.__command(self.__stats, containers=containers)

    def status(self, containers=[]):
        if not containers:
            try:
                containers = self.getDefinedServices()
            except Exception as e:
                return Fail(e)
        return self.__readStates(containers) \
            .bind(self.__status, containers=containers)

    def reset(self, time=docker.DOCKER_STOP_TIME):
        managed = docker.readManagedContainers()
        if managed.isFail():
            return managed

        return managed \
            .bind(docker.filterExistingContainers) \
            .bind(docker.readContainersStatus) \
            .bind(self.__remove, containers=managed.getOK(), time=time, force=True)

    def pull(self, containers=[], all=False):
        if all:
            try:
                containers = self.getDefinedServices()
            except Exception as e:
                return Fail(e)

        registries = self.getRegistries()

        def pullImage(container):
            image = self.getContainerImage(container)
            unpacked = docker.unpackImageString(image)
            registry = unpacked.get('registry')
            op = OK(None)
            if registry and registry in registries:
                op.then(defer(self.login, registry=registry))
            return op.then(defer(docker.pull, image=image)) \
                .then(dinfo("'%s' (%s) has been pulled." % (container, image)))

        return Try.sequence(list(map(pullImage, containers)))

    def login(self, registry):
        registries = self.getRegistries()
        config = registries.get(registry)
        if not config:
            return Fail(InvalidRegistryError("Invalid registry specified for login. It is not configured."))
        logger.info("Logging into registry: %s" % registry)
        return docker.login(registry, config.get('username'), config.get('password'), config.get('email'))

    def recreate(self, containers=[], all=False, time=docker.DOCKER_STOP_TIME):
        if all:
            try:
                containers = self.getDefinedServices()
            except Exception as e:
                return Fail(e)
        return self.__readStates(containers) \
            .bind(self.__remove, containers=containers, force=True, time=time) \
            .then(defer(self.__readStates, containers=containers)) \
            .bind(self.__start, containers=containers)

    def run(self, container, containerArgs):
        jobs = self.getDefinedJobs()
        try:
            if container not in jobs:
                return Fail(InvalidContainerError("Container %s is not a defined job." % container))
        except Exception as e:
            return Fail(e)
        return docker.run(container, containerArgs, self.getJobConfig(container), basePath=self.getBasePath(), networks=self.getNetworks())

    def excmd(self, container, cmd, tty=False, interactive=False, user=None, detach=None, privileged=None):
        return self.__readStates([container]) \
            .bind(self.__exec, container=container, cmd=cmd, tty=tty, interactive=interactive, user=user, detach=detach, privileged=privileged)

    def __stats(self, state, containers=[]):
        existing = [x for x in containers if x in state]
        return docker.stats(existing)

    def __exec(self, state, container, cmd, tty=False, interactive=False, user=None, detach=False, privileged=False):
        if container not in state:
            return Fail(InvalidContainerError("'%s' does not exist. Cannot execute command." % container))
        if not state[container].running:
            return Fail(InvalidContainerError("'%s' is not running. Cannot execute command." % container))

        return docker.execmd(container, cmd, tty, interactive, user, detach, privileged)

    def __command(self, func, containers=[], all=False, *args, **kwargs):
        if all:
            try:
                containers = self.getDefinedServices(func, all)
            except Exception as e:
                return Fail(e)
        return self.__readStates(containers) \
            .bind(func, containers=containers, *args, **kwargs)

    def __readStates(self, containers):
        return self.readOrderedContainers(containers) \
            .bind(docker.filterExistingContainers) \
            .bind(docker.readContainersStatus)

    def __status(self, state, containers=[]):
        table = []
        for container in containers:
            if container not in state:
                status = docker.ContainerStatus(container)
            else:
                status = state[container]

            row = [
                container,
                status.getCol('cid'),
                status.getCol('pid'),
                status.getCol('ip'),
                dateToAgo(status.startedat) if status.startedat else "-",
                docker.getErrorLabel(status) if not status.running else "-"
            ]
            table.append(row)
        return OK(table)

    def __create(self, state, containers=[]):
        ops = []
        for container in containers:
            if container not in state:
                op = docker.create(container, self.getContainerConfig(container), basePath=self.getBasePath(), networks=self.getNetworks()) \
                    .then(dinfo("'%s' has been created." % container))
                ops.append(op)
            else:
                logger.warn("'%s' already exists." % container)
        return Try.sequence(ops)

    def __start(self, state, containers=[]):
        ops = []
        for container in containers:
            if container not in state:
                op = docker.create(container, self.getContainerConfig(container), basePath=self.getBasePath(),  networks=self.getNetworks()) \
                    .then(defer(docker.start, container=container)) \
                    .then(dinfo("'%s' has been created and started." % container))  \
                    .then(defer(self.writePid, container=container))
                ops.append(op)
            else:
                if not state[container].running:
                    op = docker.start(container) \
                        .then(dinfo("'%s' has been started." % container)) \
                        .then(defer(self.writePid, container=container))
                    ops.append(op)
                else:
                    logger.warn("'%s' is already running." % container)
        return Try.sequence(ops)

    def __stop(self, state, containers=[], time=docker.DOCKER_STOP_TIME):
        ops = []
        for container in containers:
            if container not in state:
                logger.warn("Container '%s' does not exist." % container)
            else:
                if state[container].running:
                    op = docker.stop(container, time) \
                        .then(dinfo("'%s' has been stopped." % container)) \
                        .then(defer(self.clearPid, container=container))
                else:
                    logger.warn("'%s' is not running." % container)
        return Try.sequence(ops)

    def __remove(self, state, containers=[], force=False, time=docker.DOCKER_STOP_TIME):
        logger.debug("REMOVE %s" % containers)
        ops = []
        for container in containers:
            if container in state:
                if state[container].running:
                    if not force:
                        logger.error(
                            "'%s' is running and 'force' was not specified." % container)
                    else:
                        op = docker.stop(container, time=time) \
                            .then(defer(docker.remove, container=container)) \
                            .then(dinfo("'%s' has been stopped and removed." % container)) \
                            .then(defer(self.clearPid, container=container))
                        ops.append(op)
                else:
                    ops.append(docker.remove(container).bind(
                        dinfo("'%s' has been removed." % container)))
        return Try.sequence(ops)

    def __restart(self, state, containers=[], time=docker.DOCKER_STOP_TIME):
        ops = []
        for container in containers:
            if container not in state:
                logger.error("'%s' does not exist." % container)
            else:
                if state[container].running:
                    op = docker.stop(container, time=time) \
                        .then(defer(docker.start, container=container)) \
                        .then(dinfo("'%s' has been restarted." % container)) \
                        .then(defer(self.writePid, container=container))
                    ops.append(op)
                else:
                    ops.append(docker.start(container)
                        .then(dinfo("'%s' has been started." % container))) \
                        .then(defer(self.writePid, container=container))

        return Try.sequence(ops)

    def getPidsConf(self):
        return self.config.get('pids', {})

    def getPidsDir(self):
        dir = self.getPidsConf().get('dir', '/var/run/dockwrkr')
        return expandLocalPath(dir, self.getBasePath())

    def arePidsEnabled(self):
        return self.getPidsConf().get('enabled', False)

    def writePid(self, container):
        if not self.arePidsEnabled():
            return OK(None)

        dir = self.getPidsDir()
        pidfile = os.path.join(dir, "%s.pid" % (container))

        return Shell.makeDirectory(dir) \
            .then(defer(docker.readContainerPid, container=container)) \
            .bind(defer(Try.attempt, writeToFile, filename=pidfile))

    def clearPid(self, container):
        if not self.arePidsEnabled():
            return OK(None)

        dir = self.getPidsDir()
        pidfile = os.path.join(dir, "%s.pid" % (container))

        if os.path.isfile(pidfile):
            return Try.attempt(os.remove, pidfile)
        else:
            return OK(None)



import logging
import os
import arrow
import subprocess
import semver

from dockwrkr.monads import *
from dockwrkr.shell import Shell
from dockwrkr.utils import (ensureList, expandLocalPath, safeQuote)
from dockwrkr.exceptions import ShellCommandError, DockerError, InvalidConfigError

logger = logging.getLogger(__name__)

DOCKWRKR_LABEL_DOMAIN = 'ca.turbulent.dockwrkr'
DOCKWRKR_LABEL_DOMAIN_NETWORK = 'ca.turbulent.dockwrkr-network'

DOCKER_STOP_TIME = 10
DOCKER_CLIENT = "docker"

DOCKER_LIST_OPTIONS = [
    'add-host',
    'cap-add',
    'cap-drop',
    'device',
    'dns-opt',
    'dns-search',
    'dns',
    'env-file',
    'env',
    'expose',
    'extra-flags',
    'group-add',
    'label-file',
    'label',
    'link',
    'log-opt',
    'lxc-conf',
    'publish',
    'security-opt',
    'ulimit',
    'volume',
    'volumes-from',
    'sysctl',
]

DOCKER_MAPEQUAL_OPTIONS = [
    'env',
]

DOCKER_MAP_OPTIONS = [
    'add-host',
    'link',
    'publish',
    'volume',
]


DOCKER_SINGLE_OPTIONS = [
    'blkio-weight',
    'cgroup-parent',
    'cidfile',
    'cpu-period',
    'cpu-quota',
    'cpu-shares',
    'cpuset-cpus',
    'cpuset-mems',
    'entrypoint',
    'hostname',
    'ip',
    'ipc',
    'kernel-memory',
    'log-driver',
    'mac-address',
    'memory-reservation'
    'memory-swap',
    'memory-swapiness',
    'memory',
    'name',
    'net',
    'pid',
    'restart',
    'shm-size',
    'stop-signal',
    'user',
    'uts',
    'workdir',
]

DOCKER_BOOL_OPTIONS = [
    'disable-content-trust',
    'oom-kill-disable',
    'privileged',
    'publish-all',
    'read-only',
    'sig-proxy',
]

DOCKER_NETWORK_LIST_OPTIONS = [
    "gateway",
    "ip-range",
    "label",
    "subnet",
]

DOCKER_NETWORK_SINGLE_OPTIONS = [
    "driver",
    "ipam-driver",
]
DOCKER_NETWORK_MAP_OPTIONS = [
]

DOCKER_NETWORK_MAPEQUAL_OPTIONS = [
    "aux_address",
    "ipam-opt",
    "opt",
]
DOCKER_NETWORK_BOOL_OPTIONS = [
    "internal",
    "ipv6",
]

def dockerReadCommand(cmd, params="", shell=False, stream=False, cwd=None):
    return Shell.command("%s %s %s" % (DOCKER_CLIENT, cmd, params), shell=False) \
      .catch(onDockerError)


def dockerCallCommand(cmd, params=""):
    return Shell.call("%s %s %s" % (DOCKER_CLIENT, cmd, params))


def onDockerError(err):
    return Fail(DockerError(
        message=err.stderr,
        code=err.code,
        stdout=err.stdout,
        stderr=err.stderr,
        cmd=err.cmd))

# -- API


def readManagedContainers():
    filters = "-q -a --filter 'label=%s.managed=1' --format '{{.Label \"%s.name\"}}'" % (
        DOCKWRKR_LABEL_DOMAIN, DOCKWRKR_LABEL_DOMAIN)
    return dockerReadCommand("ps", filters) \
        .bind(parseContainerList)


def filterExistingContainers(containers):
    return readManagedContainers() \
        .map(lambda l: [x for x in l if x in containers])

def readServerVersion():
    return dockerReadCommand("version", "--format '{{.Server.Version}}'") \
        .bind(lambda c: OK(c.get('stdout')))

def readContainerExists(container):
    filter = "-q -a --filter \"label=%s.name=%s\" --format '{{.Label \"%s.name\"}}'" % (
        DOCKWRKR_LABEL_DOMAIN, safeQuote(container), DOCKWRKR_LABEL_DOMAIN)
    return dockerReadCommand("ps", filter) \
        .bind(parseContainerList) \
        .map(__listToBool)


def readNetworkExists(network):
    net_filter = '-q --filter \"name=%s\"' % (list(network.keys())[0])
    if dockerReadCommand('network ls', net_filter) \
        .bind(parseContainerList) \
            .map(__listToBool).value:
        return Fail("Network Exists")
    else:
        return OK(network)


def readContainerRunning(container):
    inspect = "--format=\"{{.State.Running}}\" %s 2> /dev/null" % (
        safeQuote(container))
    return dockerReadCommand("inspect", inspect) \
        .bind(lambda r: OK(True) if r['stdout'] == "true" else OK(False))


def readContainerGhosted(container):
    inspect = "--format=\"{{ .State.Ghost }}\" %s" % (safeQuote(container))

    def ghosted(r):
        if r.out == "true":
            logger.warning("WARNING - '%s' is ghosted." % container)
            return OK(True)
        return OK(False)
    return dockerReadCommand("inspect", inspect) \
        .bind(ghosted)


def readContainersStatus(containers=[]):

    if not containers:
        return OK({})

    inspect = "-f '{{.Name}}|{{.Id}}|{{.Config.Image}}|{{.NetworkSettings.IPAddress}}|{{range $p, $conf := .NetworkSettings.Ports}}{{if $conf}}{{$p}}->{{(index $conf 0).HostPort}}{{end}} {{end}}|{{.State.Pid}}|{{.State.StartedAt}}|{{.State.Running}}|{{.State.ExitCode}}|{{.State.Error}}' %s" % (' '.join(containers))  # noqa
    return dockerReadCommand("inspect", inspect) \
        .bind(parseContainerStatus)


def readContainerPid(container):
    inspect = "--format '{{.State.Pid}}' %s" % (safeQuote(container))
    return dockerReadCommand("inspect", inspect) \
        .bind(lambda r: OK(r['stdout'].strip()))


def create(container, config, basePath=None, networks=None):
    params = readCreateParameters(
        container, config, basePath=basePath, networks=networks)
    if params.isFail():
        return params
    return dockerCallCommand("create", params.getOK())


def createNetwork(network):
    params = readCreateNetworkParameters(network)
    if params.isFail():
        return params
    return dockerCallCommand("network create", params.getOK())


def start(container):
    return dockerReadCommand("start", container)


def stop(container, time=10):
    params = "-t %s %s" % (time, container)
    return dockerReadCommand("stop", params)


def remove(container, force=False):
    params = container
    if force:
        params = "-f %s" % container
    return dockerReadCommand("rm", params)


def pull(image):
    return dockerCallCommand("pull", image) \
        .catchError(ShellCommandError, defer(_pullLoginChain, image=image))


def _pullLoginChain(err, image):
    parts = unpackImageString(image)
    if parts.get('registry'):
        return login(parts.get('registry')).then(defer(pull, image=image))
    else:
        return Fail(err)


def unpackImageString(imageStr):
    parts = imageStr.split('/', 1)
    unpack = {}
    if len(parts) > 1:
        unpack['registry'] = parts[0]
        unpack['image'] = parts[1]
    else:
        unpack['registry'] = None
        unpack['image'] = imageStr
    return unpack


def run(container, containerArgs, config, basePath=None, networks=None):
    params = readCreateParameters(
        container, config, basePath=basePath, networks=networks, asList=True)
    if params.isFail():
        return params
    try:
        cmd = [DOCKER_CLIENT, "run", "--rm", "--interactive", "--tty"] + params.getOK() + containerArgs
        logger.debug("EXECVP - %s" % subprocess.list2cmdline(cmd))
        os.execvp(DOCKER_CLIENT, cmd)
    except Exception as ex:
        return Fail(ex)


def execmd(container, cmd, tty=False, interactive=False, user=None, detach=None, privileged=None):
    opts = []
    if tty:
        opts.append('--tty')
    if interactive:
        opts.append('--interactive')
    if user:
        opts.append('--user %s' % user)
    if detach:
        opts.append('--detach')
    if privileged:
        opts.append('--privileged')

    parts = []
    parts.append(' '.join(opts))
    parts.append(container)
    parts.append(' '.join(cmd))

    return dockerCallCommand("exec", ' '.join(parts))


def stats(containers=[]):
    opts = ['-a']

    parts = []
    parts.append(' '.join(opts))
    parts.append(' '.join(containers))

    return dockerCallCommand("stats", ' '.join(parts))


def login(registry, username=None, password=None, email=None):
    opts = []

    vers = readServerVersion()
    if vers.isFail():
        return vers
    else:
        vers = vers.getOK()

    if username:
        opts.append("-u %s" % username)
    if password:
        opts.append("-p %s" % password)
    if email and semver.match(vers, "<17.0.6"):
        opts.append("-e %s" % email)

    opts.append(registry)
    return dockerCallCommand("login", ' '.join(opts))


def logout(registry):
    return dockerCallCommand("logout", "%s" % (registry))


def readCreateNetworkParameters(network):
    network_name = list(network.keys())[0]
    network_params = network[network_name]

    cmd = []
    cmd_opt = []
    cmd_list = []
    cmd_map = []

    for confkey, confval in network_params.items():
        if confkey in DOCKER_NETWORK_MAP_OPTIONS or confkey in DOCKER_NETWORK_MAPEQUAL_OPTIONS:
            if isinstance(confval, dict):
                for ck, cv in confval.items():
                    if confkey in DOCKER_NETWORK_MAPEQUAL_OPTIONS:
                        cmd_map.append("--%s=\"%s\"=\"%s\"" %
                                       (confkey, ck, safeQuote(cv)))
                    else:
                        cmd_map.append("--%s=%s:%s" %
                                       (confkey, ck, safeQuote(cv)))
                continue
        if confkey in DOCKER_NETWORK_SINGLE_OPTIONS:
            cmd_opt.append("--%s=%s" % (confkey, safeQuote(confval)))
        elif confkey in DOCKER_NETWORK_BOOL_OPTIONS:
            if confval is not None and (confval is True or confval.lower() in ["true", "yes"] or confval == 1):
                cmd_opt.append("--%s" % confkey)
        elif confkey in DOCKER_NETWORK_LIST_OPTIONS:
            if not isinstance(confval, (str, list, tuple)):
                return Fail(InvalidConfigError(
                    "[%s] Malformed option '%s': Should be string, number or list." % (network_name, confkey)))
            confval = ensureList(confval)
            for i, lv in enumerate(confval):
                cmd_list.append("--%s=%s" % (confkey, safeQuote(lv)))
        else:
            return Fail(InvalidConfigError("[%s] Unkown option '%s'." % (network_name, confkey)))

    for part in cmd_opt:
        cmd.append(part)
    for part in cmd_list:
        cmd.append(part)
    for part in cmd_map:
        cmd.append(part)

    cmd.append("--label %s.name=\"%s\"" %
               (DOCKWRKR_LABEL_DOMAIN_NETWORK, network_name))
    cmd.append("--label %s.managed=1" % DOCKWRKR_LABEL_DOMAIN)
    cmd.append(network_name)

    cmdline = ""
    for i, part in enumerate(cmd):
        if i > 0:
            cmdline += " "
        cmdline += part

    return OK(cmdline)


def readCreateParameters(container, config, basePath=None, networks=None, asList=False):

    cconf = config.copy()

    # Autostart is only used by dockwrkr, we do not want to pass it to docker
    if 'autostart' in cconf:
        del cconf['autostart']

    if 'net' in cconf:
        if networks and cconf['net'] in networks:
            readNetworkExists({cconf['net']: networks[cconf['net']]}) \
                .bind(createNetwork)

    cconf['name'] = container

    if 'image' not in cconf:
        return Fail(InvalidConfigError("[%s] Container has no 'image' defined." % (container)))
    image = cconf['image']
    del cconf['image']

    command = []
    if 'command' in cconf:
        command = ensureList(cconf['command'])
        del cconf['command']

    extra_flags = []
    if 'extra-flags' in cconf:
        extra_flags = ensureList(cconf['extra-flags'])
        del cconf['extra-flags']

    cmd = []
    cmd_opt = []
    cmd_list = []
    cmd_map = []

    for confkey, confval in cconf.items():
        if confkey in DOCKER_MAP_OPTIONS or confkey in DOCKER_MAPEQUAL_OPTIONS:
            if isinstance(confval, dict):
                for ck, cv in confval.items():
                    if confkey in DOCKER_MAPEQUAL_OPTIONS:
                        cmd_map.append("--%s=%s=%s" %
                                       (confkey, ck, safeQuote(cv)))
                    else:
                        cmd_map.append("--%s=%s:%s" %
                                       (confkey, ck, safeQuote(cv)))
                continue
        if confkey == "volume":
            for ck in ensureList(confval):
                if ck.find(':'):
                    (path, sep, path_map) = ck.partition(':')
                    path = expandLocalPath(path, basePath=basePath)
                    cmd_map.append("--%s=%s:%s" %
                                   (confkey, safeQuote(path), safeQuote(path_map)))
            continue
        if confkey in DOCKER_SINGLE_OPTIONS:
            cmd_opt.append("--%s=%s" % (confkey, safeQuote(confval)))
        elif confkey in DOCKER_BOOL_OPTIONS:
            if confval is not None and (confval is True or confval.lower() in ["true", "yes"] or confval == 1):
                cmd_opt.append("--%s=%s" % (confkey, "true"))
            elif confval is not None and (confval is False or confval.lower() in ["false", "no"] or confval == 0):
                cmd_opt.append("--%s=%s" % (confkey, "false"))
            else:
                return Fail(InvalidConfigError("[%s] Invalid value for option '%s' : Must be true or false." % (container, confkey)))
        elif confkey in DOCKER_LIST_OPTIONS:
            if not isinstance(confval, (str, list, tuple)):
                return Fail(InvalidConfigError("[%s] Malformed option '%s': Should be string, number or list." % (container, confkey)))
            confval = ensureList(confval)
            for i, lv in enumerate(confval):
                cmd_list.append("--%s=%s" % (confkey, safeQuote(lv)))
        else:
            return Fail(InvalidConfigError("[%s] Unknown option '%s'." % (container, confkey)))

    for part in cmd_opt:
        cmd.append(part)
    for part in cmd_list:
        cmd.append(part)
    for part in cmd_map:
        cmd.append(part)
    for extra in extra_flags:
        cmd.append(extra)

    cmd.append("--label")
    cmd.append("%s.name=\"%s\"" % (DOCKWRKR_LABEL_DOMAIN, container))
    cmd.append("--label")
    cmd.append("%s.managed=1" % DOCKWRKR_LABEL_DOMAIN)
    cmd.append(image)

    for part in command:
        cmd.append(part)

    if asList:
        return OK(cmd)

    cmdline = ""
    for i, part in enumerate(cmd):
        if i > 0:
            cmdline += " "
        cmdline += part

    return OK(cmdline)

# ---- Parse command output


def parseContainerList(containers):
    return OK(containers['stdout'].strip().splitlines())


class ContainerStatus(object):
    def __init__(self, name):
        self.name = name
        self.cid = None
        self.image = None
        self.pid = None
        self.ip = None
        self.ports = None
        self.startedat = None
        self.running = None
        self.exitcode = None
        self.exiterr = None

    @staticmethod
    def fromStatusLine(line):
        parts = line.split('|')
        parts += [None] * (10 - len(parts))

        name = parts[0]
        cid = parts[1]
        image = parts[2]
        ip = parts[3]
        ports = parts[4]
        pid = parts[5]
        startedat = parts[6]
        running = parts[7]
        exitcode = parts[8]
        exiterr = parts[9]

        name = name[1:]
        status = ContainerStatus(name)
        status.cid = cid[0:12] if cid else None
        status.image = image
        status.ip = ip if ip else None
        status.pid = int(pid) if pid and pid != "0" else None
        status.running = True if running == 'true' else False
        status.startedat = arrow.get(
            startedat).timestamp if startedat and status.running else None
        status.exitcode = int(float(exitcode)) if exitcode else None
        status.exiterr = exiterr
        return status

    def getCol(self, field):
        if hasattr(self, field):
            v = getattr(self, field)
            return v if v is not None else "-"
        else:
            return "-"

    def __repr__(self):
        return "%s" % self.__dict__

def parseContainerStatus(inspect):
    statuses = {}
    for line in inspect['stdout'].strip().splitlines():
        status = ContainerStatus.fromStatusLine(line)
        statuses[status.name] = status
    return OK(statuses)

# -- Transformations


def getErrorLabel(status):
    errcode = status.exitcode
    err = status.exiterr
    if errcode == -1:
        return "%s" % err
    elif errcode == 1:
        return "General error"
    elif errcode == 126:
        return "Command invoked cannot execute"
    elif errcode == 127:
        return "Command not found"
    elif errcode == 137:
        return "SIGKILL received"
    elif errcode == 143:
        return "SIGTERM received"
    elif errcode == 0:
        return "-"
    elif errcode is not None or err is not None:
        return "Exit code %s: %s" % (errcode, err)
    else:
        return "-"


def __listToBool(l):
    return len(l) > 0



from dockwrkr import (Command)


class Run(Command):

    def getUsage(self):
        return "dockwrkr run JOBCONTAINER [ARGS...]"

    def getHelpTitle(self):
        return "Run the specified job container"

    def execute(self, args=None):
        # Do NOT allow interspersed options!
        self.input = args
        (self.options, self.args) = self.parseShellInput(False)
        return self.main()

    def main(self):
        if not len(self.args) > 0:
            return self.exitError("Please provide a job container to run.")
        container = self.args[0]
        container_args = self.args[1:]
        return self.core.run(container, container_args) \
            .catch(self.exitError)



import unittest
import os
import tempfile
import tests
from contextlib import contextmanager

from dockwrkr.core import Core
from dockwrkr.shell import Shell
from dockwrkr.monads import *


class TestCore(tests.TestBase):

    def setUp(self):
        self.basePath = self.addTemporaryDir()

    def tearDown(self):
        ''' tearDown the test '''
        if self.basePath:
            Shell.nukeDirectory(self.basePath).catch(TestCore.raiser)

    def setUpDefaultCore(self):
        self.core = Core()
        return self.core

    def testInitialize(self):
        Shell.call("cp %s %s" % ("tests/dockwrkr-1.yml",
                                 os.path.join(self.basePath, "dockwrkr.yml")))
        with self.pushd(self.basePath):
            core = Core()
            self.assertIsInstance(core.initialize(), OK)

            self.assertIsInstance(core.reset(time=0), OK)
            self.assertIsInstance(core.start(containers=['hello1']), OK)
            self.assertIsInstance(core.start(containers=['hello2']), OK)
            self.assertIsInstance(core.start(containers=['hello3']), OK)
            self.assertIsInstance(core.start(containers=['hello4']), OK)
            self.assertIsInstance(core.stop(all=True, time=0), OK)
            self.assertIsInstance(core.start(all=True), OK)

            # Test hello4 has no PID
            self.assertEqual('-', core.status(containers=['hello4']).getOK()[0][2])

            self.assertIsInstance(core.pull(all=True), OK)
            self.assertIsInstance(core.excmd(
                container='hello1', cmd=['ls', '-al']), OK)
            self.assertIsInstance(core.reset(time=0), OK)

    def testPids(self):
        Shell.call("cp %s %s" % ("tests/dockwrkr-2.yml",
                                 os.path.join(self.basePath, "dockwrkr.yml")))
        with self.pushd(self.basePath):
            core = Core()
            self.assertIsInstance(core.initialize(), OK)

            pdir = os.path.join(self.basePath, "pids")

            self.assertIsInstance(core.reset(time=0), OK)
            self.assertIsInstance(core.start(containers=['hello1']), OK)
            self.assertTrue(os.path.isfile(os.path.join(pdir, "hello1.pid")))
            self.assertIsInstance(core.start(containers=['hello2']), OK)
            self.assertTrue(os.path.isfile(os.path.join(pdir, "hello2.pid")))
            self.assertIsInstance(core.start(containers=['hello3']), OK)
            self.assertTrue(os.path.isfile(os.path.join(pdir, "hello3.pid")))
            self.assertIsInstance(core.stop(all=True, time=0), OK)
            self.assertFalse(os.path.isfile(os.path.join(pdir, "hello1.pid")))
            self.assertFalse(os.path.isfile(os.path.join(pdir, "hello2.pid")))
            self.assertFalse(os.path.isfile(os.path.join(pdir, "hello3.pid")))

    def testRes(self):
        Shell.call("cp %s %s" % ("tests/dockwrkr-2.yml",
                                 os.path.join(self.basePath, "dockwrkr.yml")))
        with self.pushd(self.basePath):
            core = Core()
            self.assertIsInstance(core.initialize(), OK)

            self.assertIsInstance(core.reset(time=0), OK)
            self.assertIsInstance(core.start(containers=['hello1']), OK)
            self.assertIsInstance(core.restart(
                containers=['hello1'], time=0), OK)
            self.assertIsInstance(core.recreate(
                containers=['hello1'], time=0), OK)
            self.assertIsInstance(core.reset(time=0), OK)

    def testJob(self):
        Shell.call("cp %s %s" % ("tests/dockwrkr-2.yml",
            os.path.join(self.basePath, "dockwrkr.yml")))

        with self.pushd(self.basePath):
            core = Core()
            self.assertIsInstance(core.initialize(), OK)
            core.run('hellojob', ['Executing dockwrkr job: OK!'])

    @contextmanager
    def pushd(self, newDir):
        previousDir = os.getcwd()
        os.chdir(newDir)
        yield
        os.chdir(previousDir)



import os
import sys
import logging
import shutil
import shlex
import subprocess
from subprocess import call, Popen, check_output, CalledProcessError
from dockwrkr.monads import *
from dockwrkr.exceptions import (
    FileSystemError, ShellCommandError, UserInterruptError)
from threading import Thread

# pylint: disable=W0232

logger = logging.getLogger(__name__)


class Shell(object):

    @staticmethod
    def printConfirm(msg, assumeYes=False):
        if assumeYes:
            return OK(True)

        logger.info(msg)
        try:
            res = input('Proceed? [N/y] ')
            if res.lower().startswith('y'):
                logger.info('... proceeding')
                return OK(True)
            return Fail(UserInterruptError(message="User interrupted."))
        except KeyboardInterrupt as err:
            return Fail(UserInterruptError(message="User interrupted."))

    @staticmethod
    def call(cmd, cwd=None, shell=True):
        logger.debug("COMMAND[%s]: %s", cwd, cmd)
        try:
            returncode = call(cmd, shell=shell, cwd=cwd)
            if returncode == 0:
                return OK(None)
            else:
                return Fail(ShellCommandError(code=returncode))
        except CalledProcessError as err:
            return Fail(ShellCommandError(code=err.returncode, message=err.output, stdout=err.output))
        except KeyboardInterrupt as err:
            logger.info("CTRL-C Received...Exiting.")
            return Fail(UserInterruptError(message="User interrupted."))

    @staticmethod
    def command(cmd, shell=False, cwd=None):
        logger.debug("COMMAND: %s", cmd)
        try:
            out = check_output(shlex.split(cmd), shell=shell, cwd=cwd)
            return OK({'code': 0, 'stdout': out.decode().strip(), 'stderr': ''})
        except CalledProcessError as err:
            return Fail(ShellCommandError(code=err.returncode, message=err.output, stdout=err.output))

    @staticmethod
    def procCommand(cmd, cwd=None, shell=False):
        try:
            logger.debug("COMMAND: %s", cmd)
            proc = Popen(shlex.split(cmd), shell=shell,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
            pout, perr = proc.communicate()
            pout = pout.decode()
            perr = perr.decode()
            if proc.returncode == 0:
                return OK({"code": proc.returncode, "stdout": pout, "stderr": perr})
            else:
                return Fail(ShellCommandError(code=proc.returncode, message=pout, stdout=pout, stderr=perr))
        except KeyboardInterrupt:
            logger.info("CTRL-C Received...Exiting.")
            return Fail(UserInterruptError(message="User interrupted."))

    @staticmethod
    def streamCommand(cmd, cwd=None, shell=False, stream=False):
        try:
            logger.debug("COMMAND: %s", cmd)
            proc = Popen(shlex.split(cmd), shell=shell,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
            stdout = b''
            stderr = b''
            while True:
                d = proc.stdout.read(1)
                if d != '':
                    stdout += d
                if stream:
                    sys.stdout.write(d.encode())
                    sys.stdout.flush()

                de = proc.stderr.read(1)
                if de != '':
                    stderr += de
                if stream:
                    sys.stderr.write(de.encode())
                    sys.stderr.flush()

                if d == '' and de == '' and proc.poll() is not None:
                    break

            stdout = stdout.decode()
            stderr = stderr.decode()
            if proc.returncode == 0:
                return OK({"code": proc.returncode, "stdout": stdout, "stderr": stderr})
            else:
                return Fail(ShellCommandError(code=proc.returncode, message=stdout, stdout=stdout, stderr=stderr))
        except OSError as err:
            err.strerror = "Error running '%s': %s" % (cmd, err.strerror)
            return Fail(err)
        except KeyboardInterrupt:
            logger.info("CTRL-C Received...Exiting.")
            return Fail(UserInterruptError(message="User interrupted."))

    @staticmethod
    def chmod(path, mode):
        try:
            os.chmod(path, mode)
            return OK(path)
        except Exception as err:
            return Fail(ShellCommandError(code=1, message="Failed to chmod %s: %s" % (path, err)))

    @staticmethod
    def makeDirectory(path, mode=0o750):
        if not os.path.exists(path):
            try:
                os.makedirs(path, mode)
                return OK(path)
            except Exception as err:
                return Fail(ShellCommandError(code=1, message="Failed to create %s: %s" % (path, err)))
        return OK(path)

    @staticmethod
    def copyFile(src, dst):
        return Try.attempt(lambda: shutil.copy(src, dst))

    @staticmethod
    def pathExists(path):
        return OK(path) if os.path.exists(path) else Fail(FileSystemError("Path %s does not exist." % path))

    @staticmethod
    def rmFile(path):
        if os.path.isdir(path):
            return Fail(FileSystemError("%s is a directory.") % path)
        return OK(os.remove(path))

    @staticmethod
    def nukeDirectory(path):
        try:
            if path and path is not '/' and path is not 'C:\\':
                shutil.rmtree(path)
            return OK(None)
        except Exception as err:
            return Fail(ShellCommandError(code=1, message="Failed to rmtree: %s: %s" % (path, err)))



from dockwrkr import (Command)
from dockwrkr.monads import (Try)


class Login(Command):

    def getUsage(self):
        return "dockwrkr login [REGISTRY...]"

    def getHelpTitle(self):
        return "Perform docker login using credentials in dockwrkr.yml"

    def main(self):
        if self.args:
            registries = self.args
        else:
            registries = self.core.getRegistries()
        results = [self.core.login(registry) for registry in registries]
        return Try.sequence(results).catch(self.exitError)



from setuptools import setup, find_packages

with open('README.rst') as f:
    readme = f.read()

with open('dockwrkr/_version.py') as versionFile:
    exec(versionFile.read())

install_requires = ['setuptools', 'PyYAML', 'arrow', 'tabulate', 'semver>=2.7']

setup(name='dockwrkr',
      version=__version__,
      author='Turbulent inc.',
      author_email='oss@turbulent.ca',
      license='Apache License 2.0',
      long_description=readme,
      description='dockwrkr - docker container launch wrapper',
      long_description_content_type='text/markdown',
      install_requires=install_requires,
      test_suite='tests',
      packages=find_packages(),
      python_requires='>=3',
      entry_points={
          'console_scripts': [
              'dockwrkr = dockwrkr.cli:cli',
          ],
      })
